/* 
 * test_prog/csparsetest/csparsetest1a.c                *** Node Program ***
 * Csparse sparse matrix library tester component.
 * by Anthony Skjellum and Alvin P. Leung
 *
 * Anthony Skjellum
 * Caltech Chemical Engineering
 * 210-41 1201 E. California Blvd.
 * Pasadena, CA 91125
 * (818) 356-4634
 * ARPAnet: tony@juliet.caltech.edu; BITNET: tony@citjulie.bitnet
 * 
 * .   .   .   .   .   .   .   .   .
 * 
 * 
 * (C) 1989, 1990, 1991 California Institute of Technology, Inc.
 * (C) 1991 Regents of the University of California
 * 
 * Zipcode version history
 * 
 * created:     **-Sep-89 {APL}
 * last update: 06-May-90 based on ... 28-Feb-90 based on 27-Feb-90 {AS}
 * 
 * 06-May-90:   in stat loop, increase debugging # of memory tests...
 * 14-Mar-90:   in statistics dump to host, include sum_squared tag, which
 *              is the sum of the squares of values in the distribution.  
 *              useful for extending or combining statistics later...
 *              comparable extension in csparsetest1.c to accept new data.
 *
 * 12-Mar-90:   many things have been added since march 5.  Included today:
 *              led blinking lights for debugging;  memory consumption 
 *              monitoring on a global basis inside the main loop.
 * 11-Mar-90:   synchronized debugging statement in main test loop.
 *              reduction of the amount of output.
 * 05-Mar-90:   upgrade for sband generation letters/call
 * 28-Feb-90:   more of the same...
 * 27-Feb-90:   basically, rewrite this entire code for the new sparse matrix
 *              library, and improve its behavior insofar as timing accuracy.
 *
 *              First, re-do the code so that the host transmits the sparse
 *              matrix directly, instead of a dense representation of the 
 *              sparse matrix.
 */


#include <csparse.h>
#include <cstats.h>
#include <csparsetest.h>

#include <sys/time.h>
#include <sys/resource.h>

static struct rlimit rlp;

#if 0
#define RLIM    (getrlimit(RLIMIT_DATA,&rlp),rlp.rlim_max),_res_malloc_quant
#endif
#define SIMPLE_RLIM     (getrlimit(RLIMIT_DATA,&rlp),rlp.rlim_max)

#if 0
#define DEBUG0        1
#define DEBUG1        1
#define DEBUG2        1
#define DEBUG3        1
#define DEBUG4        1
#define DEBUG5        1
#define DEBUG6        1
#define DEBUG7        1
#define DEBUG8        1
#define DEBUG9        1

#define DEBUG_STAT_LOOP 1
#define DEBUG_STAT_LOOP_MEM_DISPL 0
#define DEBUG7dot1    1

#define DEBUG_COMBINE 1
#define MEMORY_DISPL 0

#define LED_DISPL 0
#endif

/* debugging controls: */
#define STASIS

#define NSTATS           8

#define DEBUG_STATS_IN_MAIN
main (argc, argv)
int argc;
char *argv[];
{
    STASIS int            i EQ_ZERO;

    STASIS ZIP_MAILER     *Node_Mailer EQ_NULL;
    STASIS ZIP_MAILER     *Mailer EQ_NULL;
    STASIS int             p;
    STASIS int             q;
    STASIS int             P;
    STASIS int             Q;
    STASIS int             Max_Grid_Row EQ_ZERO;
    STASIS int             Max_Grid_Col EQ_ZERO;
    STASIS int             Grid_Row EQ_ZERO;
    STASIS int             Grid_Col EQ_ZERO;
    STASIS G2_POBOX        Recv_POBox;
    STASIS G2_POBOX        Send_POBox;

    STASIS char           *Letter EQ_NULL;
    STASIS long            Letter_Size EQ_ZERO;
    STASIS int             Letter_Type EQ_ZERO;

    STASIS long           *Header EQ_NULL;
    STASIS long           *Dis_Header EQ_NULL;
    STASIS double         *Real_Objs EQ_NULL;

    STASIS short           Row_Dis_Type EQ_NULL;
    STASIS short           Col_Dis_Type EQ_NULL;
    STASIS long           *Mat_Data_Header EQ_NULL;
    STASIS long            Matrix_Row_Size EQ_ZERO;
    STASIS long            Matrix_Col_Size EQ_ZERO;
    STASIS long            Repeat EQ_ZERO;
    STASIS long            Dump_to_Host_Freq EQ_ZERO, Dump_Count EQ_ZERO;
    STASIS long            Strategy EQ_ZERO;
    STASIS int             Regen_Flag EQ_ZERO;

    STASIS Cdistrib       *Row_Dis EQ_NULL;
    STASIS void          (*Row_Distribution)() EQ_NULL;
    STASIS int           (*Row_Inverse_Distribution)() EQ_NULL;
    STASIS int           (*Row_Distribution_limit)() EQ_NULL;
    STASIS int           (*Row_Distribution_init)() EQ_NULL;
    STASIS index           Local_Max_Row EQ_ZERO;
    STASIS int             Row_Extra_param1 EQ_ZERO;
    STASIS int             Row_Extra_param2 EQ_ZERO;
    STASIS void           *Row_Extra EQ_NULL;

    STASIS Cdistrib       *Col_Dis EQ_NULL;
    STASIS void          (*Col_Distribution)() EQ_NULL;
    STASIS int           (*Col_Inverse_Distribution)() EQ_NULL;
    STASIS int           (*Col_Distribution_limit)() EQ_NULL;
    STASIS int           (*Col_Distribution_init)() EQ_NULL;
    STASIS index           Local_Max_Col EQ_ZERO;
    STASIS int             Col_Extra_param1 EQ_ZERO;
    STASIS int             Col_Extra_param2 EQ_ZERO;
    STASIS void           *Col_Extra EQ_NULL;

    STASIS Cvector        *Right_Hand_Side EQ_NULL,
                          *Extra_Right_Hand_Side EQ_NULL;
    STASIS double         *RHS EQ_NULL, *Extra_RHS EQ_NULL;
    STASIS Cvector        *Solution EQ_NULL;
    STASIS double         *Sol EQ_NULL;

    STASIS int             Counter EQ_ZERO;
    STASIS CMdistrib      *Matrix_Dis EQ_NULL;
    STASIS Dmatrix        *Csparse_Matrix EQ_NULL;
    STASIS Smatrix        *Matrix EQ_NULL;
    STASIS int             Local_Row_Index EQ_ZERO;
    STASIS int             Local_Col_Index EQ_ZERO;
    STASIS Dlu_info       *Matrix_LU_Info EQ_NULL;
    STASIS int             Global_Col_Index EQ_ZERO;
    STASIS int             Time_Start EQ_ZERO;
    STASIS int             Time_Stop EQ_ZERO;
    STASIS int             Size  EQ_ZERO;
    STASIS double          Value EQ_ZERO;

    STASIS DXpivot_group   *pivgrp EQ_NULL;

    STASIS short           done EQ_ZERO;
    STASIS int             serror EQ_ZERO;
    STASIS int             G2_pm EQ_ZERO;

    STASIS long n_longs EQ_ZERO, n_objs EQ_ZERO;
    STASIS double *Real_objs;

    STASIS LOCAL_REAL_VALUES *matrix_data_stack_top = NULL,
                             *matrix_data_stack = NULL,
                             *matrix_data_stack_entry  = NULL,
                             *matrix_data_stack_entry1 = NULL;

    STASIS MK_MATRIX_ENTRY *mk_matrix_stack_top = NULL, *mk_matrix_stack = NULL,
                           *matrix_entry;

    SBAND_EXTRA *sband_extra_item;
    DATA_EXTRA  *data_extra_item;


    /* matrix generation variables: */
    STASIS int Upper_Band_Width, Lower_Band_Width, skew;
    STASIS double scale, sp_scale, Fill_Percent;

    /* context control: */
    STASIS int context;

    /* statistics: */
    STASIS Cstats *mean_stats;
    STASIS STATS  *max_stats, *min_stats;

    STASIS double *mean_time, *max_time, *min_time;

    STASIS int seed;

#if DEBUG4
    STASIS int Global_Row_Index;
#endif

#if DEBUG_COMBINE
    STASIS int one_int;
#endif
#if MEMORY_DISPL
    int my_free_memory;
    int imin_max_avg_where_co();
    int min_max_avg[10];
#endif

#if MEMORY_DISPL
    my_free_memory = SIMPLE_RLIM;
#endif

#if 0
#if MEMORY_DISPL
    print("csparsetest1a -- start: my free memory = %d",my_free_memory);
#else
#if DEBUG1
    print("csparsetest1a -- start");
#endif
#endif
#endif
 
    if (Zip_init() == NULL)
    {
       print("error - zip_init()");
       exit(1);
    }

#if DEBUG1
    print("csparsetest1a -- past zip_init");
#endif
 
    if ((Mailer = g2_grid_open(&Max_Grid_Row, &Max_Grid_Col, NULL)) == NULL)
    {
       print("error - g2_grid_open()");
       exit(1);
    }
#if DEBUG1
    print("csparsetest1a -- past g2_grid_open()");
#endif
 
    Zip_push(Mailer);

    G2_pq(Grid_Row, Grid_Col);
 
    G2_pm = (!Grid_Row) && (!Grid_Col);

#if MEMORY_DISPL
    min_max_avg[0] = my_free_memory;    /* min slot */
    min_max_avg[1] = Grid_Row;
    min_max_avg[2] = Grid_Col;
    min_max_avg[3] = 1;                 /* multiplicity of min */

    min_max_avg[4] = my_free_memory;    /* max slot */
    min_max_avg[5] = Grid_Row;
    min_max_avg[6] = Grid_Col;
    min_max_avg[7] = 1;                 /* multiplicity of maximum */

    min_max_avg[8] = my_free_memory;    /* sum */
    min_max_avg[9] = 1;                 /* count */

    G2_combine(min_max_avg, imin_max_avg_where_co, 10*sizeof(int), 1);

    if(G2_pm)
    {
        print("csparsetest1a -- Checkpoint 1; ensemble memory data:");
        print("min = %5d, @(p=%d,q=%d); max = %5d @(p=%d,q=%d); `avg' = %g",
                min_max_avg[0], min_max_avg[1], min_max_avg[2],
                min_max_avg[4], min_max_avg[5], min_max_avg[6],
                (double)min_max_avg[8] / (double)min_max_avg[9]);
        print("min. multiplicity = %5d, max. multiplicity = %5d",
                min_max_avg[3], min_max_avg[7]);
    }
#else
#if DEBUG1
    G2_sync();
    if(G2_pm)
        print("csparsetest1a -- all at Checkpoint 1");
#endif
#endif

#if DEBUG_COMBINE
    one_int = 1;
    G2_combine(&one_int, iadd_co, sizeof(int), 1);
    print("TEST: Combined one's over PQ proc. = %d yields %d", 
                Max_Grid_Row * Max_Grid_Col, one_int);
    one_int = 1;
    G2_row_combine(&one_int, iadd_co, sizeof(int), 1);
    print("TEST: Combined one's over rows, Q = %d yields %d", Max_Grid_Col, one_int); one_int = Grid_Col;
    G2_row_combine(&one_int, iadd_co, sizeof(int), 1);
    print("TEST: Sum of q's by row combine, should be %d, is %d", 
                (Max_Grid_Col * (Max_Grid_Col - 1))/2, one_int);
    one_int = 1;
    G2_col_combine(&one_int, iadd_co, sizeof(int), 1);
    print("TEST: Combined one's over cols, P = %d yields %d", Max_Grid_Row, one_int);
    one_int = Grid_Row;
    G2_col_combine(&one_int, iadd_co, sizeof(int), 1);
    print("TEST: Sum of p's by col combine, should be %d, is %d",
                (Max_Grid_Row * (Max_Grid_Row - 1))/2, one_int);

    /* exit(0); */      /* enough... */
#endif
 
#if DEBUG1
    print("Max Nodes (P=%d,Q=%d), My Node (p=%d, q=%d)", Max_Grid_Row,
          Max_Grid_Col, Grid_Row, Grid_Col);
#endif

    Send_POBox.d1 = Grid_Row;
    Send_POBox.d2 = Grid_Col;
 
    if(G2_pm)
    {
#if DEBUG1
        print("I am G2 PM; waiting for letter from HOST");
#endif

        Letter      = (char *) G2_recvb(-1, -1);
        Header      = (long *)Letter;
        Letter_Size = Header[1];
    }
    else
    {
        Letter_Size = 0;
        Letter      = NULL;
    }

    /* fanout: */
    serror = G2_fanout(&Letter, &Letter_Size, 0, 0);

#if DEBUG1
    print("Made it past G2_fanout(): Letter = %lXX, Letter_Size = %d, serror = %d", Letter, Letter_Size, serror);
#endif
        
    Header = (long *)Letter;

#if DEBUG2
    print("received initial letter from HOST; Header = %lxX", Header);
#endif

    if(*Header != INTRO_LETTER)
    {
        print("csparsetest1a: anticipated INTRO_LETTER=%ld, received %ld",
                 (long)INTRO_LETTER, *Header);
        
        exit(1);
    }

    Dis_Header   = Header + 4;
    Row_Dis_Type = (short)Dis_Header[0];
    Col_Dis_Type = (short)Dis_Header[1]; 

#if DEBUG2
    print("Row_Dis_Type = %d, Col_Dis_Type = %d", Row_Dis_Type, Col_Dis_Type);
#endif

    Mat_Data_Header = (Dis_Header + 2);
    Matrix_Row_Size = Mat_Data_Header[0];
    Matrix_Col_Size = Mat_Data_Header[1];

    /* data for one and two-parameter distributions */
    Row_Extra_param1       = (int) Mat_Data_Header[2]; /* "B" */
    Row_Extra_param2       = (int) Mat_Data_Header[3]; /* "S" */
    Col_Extra_param1       = (int) Mat_Data_Header[4]; /* "B" */
    Col_Extra_param2       = (int) Mat_Data_Header[5]; /* "S" */

#if DEBUG2  
    print("Row_Extra: Par1 = %d, Par2 = %d; Col_Extra: Par1 = %d Par2 = %d",
        Row_Extra_param1, Row_Extra_param2, Col_Extra_param1, Col_Extra_param2);
#endif

    Repeat            = Mat_Data_Header[6];
    Strategy          = Mat_Data_Header[7];
    Dump_to_Host_Freq = Mat_Data_Header[8];
    Regen_Flag        = (int)Mat_Data_Header[9];

#if DEBUG2
    print("Matrix_Row_Size = %d, Matrix_Col_Size = %d, Row_Extra = %d, Col_Extra = %d", Matrix_Row_Size, Matrix_Col_Size, Row_Extra, Col_Extra);
#endif
    if(G2_pm) Zip_free(Letter); else free(Letter);
    Letter = NULL;
    
#if DEBUG1
    print("node(p=%d,q=%d): csparsetest1a -- Checkpoint 2", Grid_Row, Grid_Col);
#endif

    /* create row distribution: */
    Row_Dis = new_Cdistrib(Mailer, CDISTRI_ROW, Row_Dis_Type, NULL);
    Row_Distribution         = Row_Dis -> mu;
    Row_Inverse_Distribution = Row_Dis -> mu_i;
    Row_Distribution_limit   = Row_Dis -> mu_lim;
    Row_Distribution_init    = Row_Dis -> mu_init;

    /* initialize the distribution accordingly: */
    switch(Row_Dis_Type)
    {
        case CDISTRI_BLINEAR:
        case CDISTRI_BSCATTERED:
        case CDISTRI_GBLINEAR:
        case CDISTRI_GBSCATTERED:
                Row_Extra = (void *)NEW_CLEAR_ITEM(int ,1);
                *(int *)Row_Extra = Row_Extra_param1;
                break;
        case CDISTRI_G1PARAM:
                Row_Extra = (void *)NEW_CLEAR_ITEM(
                                CDISTRI_GENL_1PARAM_EXTRA, 1);
                Row_Distribution_init(Row_Extra,
                        Row_Extra_param2, Max_Grid_Row, Matrix_Row_Size);
                break;
        case CDISTRI_G2PARAM:
                Row_Extra = (void *)NEW_CLEAR_ITEM(
                                CDISTRI_GENL_2PARAM_EXTRA, 1);
                Row_Distribution_init(Row_Extra, Row_Extra_param1,
                        Row_Extra_param2, Max_Grid_Row, Matrix_Row_Size);
                break;
        default:
                break;
    }
    if(Row_Dis -> mu_extra != NULL)   /* make distribution use Row_Extra */
    {
        Free(Row_Dis -> mu_extra);
        Row_Dis -> mu_extra = Row_Extra;
    }

    Local_Max_Row = (*Row_Distribution_limit)(Grid_Row, Max_Grid_Row,
                                              Matrix_Row_Size, Row_Extra);

#if DEBUG1
    print("node(p=%d,q=%d): csparsetest1a -- Checkpoint 2.1",
                Grid_Row, Grid_Col);
#endif

    /* create column distribution: */
    Col_Dis = new_Cdistrib(Mailer, CDISTRI_COLUMN, Col_Dis_Type, NULL);
    Col_Distribution         = Col_Dis -> mu;
    Col_Inverse_Distribution = Col_Dis -> mu_i;
    Col_Distribution_limit   = Col_Dis -> mu_lim;
    Col_Distribution_init    = Col_Dis -> mu_init;

    /* initialize the distribution accordingly: */
    switch(Col_Dis_Type)
    {
        case CDISTRI_BLINEAR:
        case CDISTRI_BSCATTERED:
        case CDISTRI_GBLINEAR:
        case CDISTRI_GBSCATTERED:
                Col_Extra = (void *)NEW_CLEAR_ITEM(int ,1);
                *(int *)Col_Extra = Col_Extra_param1;
                break;
        case CDISTRI_G1PARAM:
                Col_Extra = (void *)NEW_CLEAR_ITEM(
                                    CDISTRI_GENL_1PARAM_EXTRA, 1);
                Col_Distribution_init(Col_Extra,
                        Col_Extra_param2, Max_Grid_Col, Matrix_Col_Size);
                break;
        case CDISTRI_G2PARAM:
                Col_Extra = (void *)NEW_CLEAR_ITEM(
                                    CDISTRI_GENL_2PARAM_EXTRA, 1);
                Col_Distribution_init(Col_Extra, Col_Extra_param1,
                        Col_Extra_param2, Max_Grid_Col, Matrix_Col_Size);
                break;
        default:
                break;
    }
    if(Col_Dis -> mu_extra != NULL)     /* make distribution use Col_Extra */
    {
        Free(Col_Dis -> mu_extra);
        Col_Dis -> mu_extra = Col_Extra;
    }

    Local_Max_Col = (*Col_Distribution_limit)(Grid_Col, Max_Grid_Col,
                                              Matrix_Col_Size, Col_Extra);
 
#if DEBUG1
    print("node(p=%d,q=%d): csparsetest1a -- Checkpoint 2.2",
                Grid_Row, Grid_Col);
#endif
#if DEBUG3
    print("node(p=%d,q=%d): Local_Max_Row = %d, Local_Max_Col = %d",
          Grid_Row, Grid_Col, Local_Max_Row, Local_Max_Col);
#endif
/* needed in case index sets are used (not done presently): */
#if 0
    set_index_parameters(max(Matrix_Row_Size, Matrix_Col_Size),
                      2 * max(Matrix_Row_Size, Matrix_Col_Size));
#endif
    set_index_parameters(max(Local_Max_Row, Local_Max_Col),
                        2 * max(Local_Max_Row, Local_Max_Col));

    /* create the right hand size vector structure + extra copy */

    Right_Hand_Side       = new_Cvector(Row_Dis, Matrix_Row_Size);
    Extra_Right_Hand_Side = next_Cvector(Right_Hand_Side);
    RHS                   = CMATV_BASE_PTR(Right_Hand_Side);
    Extra_RHS             = CMATV_BASE_PTR(Extra_Right_Hand_Side);

#if DEBUG3
    print("made it past RHS Cvector allocations...");
#endif

    /* create the solution vector: */
    Solution              = new_Cvector(Col_Dis, Matrix_Col_Size);
    Sol                   = CMATV_BASE_PTR(Solution);

#if DEBUG3
    print("made it past Solution Cvector allocation...");
#endif

    /* create the matrix distribution structure */

    Matrix_Dis = new_CMdistrib(Row_Dis, Col_Dis);

    /* create the matrix data structure */

    Csparse_Matrix = new_Dmatrix(Matrix_Dis, Matrix_Row_Size, Matrix_Col_Size);
 
    /* reference the matrix from the structure */

    Matrix = Csparse_Matrix -> Local_Matrix;
 
#if DEBUG1
    print("node(p=%d,q=%d): csparsetest1a -- Checkpoint 3", Grid_Row, Grid_Col);
#endif
#if MEMORY_DISPL
    memory_displ(Grid_Row, Grid_Col);
#endif

    /* create the pivoting structure: */
    pivgrp = new_DXpivot_group(Strategy, CSPAR_PIV_IDX_NSCALE);

    /* create the LU information structure: */
    Matrix_LU_Info = new_Dlu_info(Csparse_Matrix, pivgrp);

    /* structures to hold the received and/or generated matrix data: */
    matrix_data_stack_top = matrix_data_stack = NULL;

    for(done = FALSE; !done; )
    {
#if DEBUG9
        print("top of reactive loop...");
#endif

        /* wait for letter: */
        Letter      = (char *) G2_recvb(G2_HOST_p, G2_HOST_q);
        Header      = (long *)Letter;
        Letter_Type = (int)Header[0];
        Letter_Size = Header[1];
        n_longs     = Header[2];
        n_objs      = Header[3];

        Real_objs   = (double *)(Header + n_longs);

#if DEBUG9
        print("Received letter of type = %d; size = %ld" , Letter_Type,
                        Letter_Size);
        print("n_longs = %ld, n_objs = %ld",n_longs, n_objs);
#endif

        switch(Letter_Type)
        {
            case SEED_RANDOM_NUMBER_GENERATOR_LETTER:
                /* seed the random number generator: */
                seed = (int)Header[4];

#if DEBUG4
                print("received seed = %d", seed);
#endif

                SRAND(seed);
                break;

            case MATRIX_GEN_SBAND_LETTER:
#if DEBUG4
                print("start of MATRIX_GEN_SBAND_LETTER switch-clause");
#endif

                Upper_Band_Width = (int)Header[4];
                Lower_Band_Width = (int)Header[5];
                skew             = (int)Header[6];
                context          = (int)Header[7];

                Fill_Percent     = Real_objs[0];
                scale            = Real_objs[1];
                sp_scale         = Real_objs[2];

                /* add the work to the matrix-make stack: */
                sband_extra_item = NEW_CLEAR_ITEM(SBAND_EXTRA, 1);
                sband_extra_item -> context = context;
                sband_extra_item -> Upper_Band_Width = Upper_Band_Width;
                sband_extra_item -> Lower_Band_Width = Lower_Band_Width;
                sband_extra_item -> skew = skew;
                sband_extra_item -> Fill_Percent = Fill_Percent;
                sband_extra_item -> scale = scale;
                sband_extra_item -> sp_scale = sp_scale;

                matrix_entry = NEW_CLEAR_ITEM(MK_MATRIX_ENTRY, 1);
                matrix_entry -> gen_class = GEN_CLASS_SBAND;
                matrix_entry -> context   = context;
                matrix_entry -> extra     = (void *)sband_extra_item;

                add_to_mk_matrix(matrix_entry,
                        &mk_matrix_stack_top, &mk_matrix_stack);

#if 0
                /* old way... */
                gen_sband_random_Dmatrix_data(Csparse_Matrix,
                                &matrix_data_stack_top, &matrix_data_stack,
                                Upper_Band_Width, Lower_Band_Width, skew,
                                Fill_Percent, scale, sp_scale, context);
#endif

#if DEBUG9
                print("end of MATRIX_GEN_SBAND_LETTER switch-clause");
#endif
                break;

            case MATRIX_INTRO_LETTER:
#if DEBUG4
                print("MATRIX_INTRO_LETTER...");
#endif
                context          = (int)Header[4];
                scale            = Real_Objs[0];

                /* copy to hierarchical data: */
                data_extra_item = NEW_CLEAR_ITEM(DATA_EXTRA, 1);
                data_extra_item -> scale = scale;

                matrix_entry = NEW_CLEAR_ITEM(MK_MATRIX_ENTRY, 1);
                matrix_entry -> gen_class = GEN_CLASS_DATA;
                matrix_entry -> context   = context;
                matrix_entry -> extra     = (void *)data_extra_item;

                add_to_mk_matrix(matrix_entry,
                        &mk_matrix_stack_top, &mk_matrix_stack);

                break;

            case MATRIX_DATA_LETTER:
#if DEBUG4
                print("MATRIX_DATA_LETTER...");
#endif

#if DEBUG4
                print("matrix_data_stack_entry = %lxX",matrix_data_stack_entry);
#endif
                context          = (int)Header[4];

                matrix_data_stack_entry = 
                    NEW_CLEAR_ITEM(LOCAL_REAL_VALUES, 1);

                matrix_data_stack_entry -> context = context;

#if 0
                if
                (
                    (matrix_data_stack_entry == NULL) ||
                    (context != matrix_data_stack_entry -> context)
                )
                {
                   print("Fatal error: context mismatch in matrix handling.");
                   exit(1);
                }
#endif

                matrix_data_stack_entry -> n_values = n_objs;

                matrix_data_stack_entry -> values   =
                    NEW_CLEAR_ITEM(LOCAL_REAL_VALUE, n_objs);

#if DEBUG4
                print("MATRIX_DATA_LETTER...checkpoint #1");
                print("n_objs=%ld, sizeof(LOCAL_REAL_VALUE) = %d",
                        n_objs, sizeof(LOCAL_REAL_VALUE));
#endif
                /* copy in data: */
                bcopy(Header + n_longs, matrix_data_stack_entry -> values,
                        (int)(n_objs * sizeof(LOCAL_REAL_VALUE)));

#if DEBUG4
                print("MATRIX_DATA_LETTER...checkpoint #2");
#endif

                matrix_data_stack_entry1 =
                         NEW_CLEAR_ITEM(LOCAL_REAL_VALUES, 1);
                bcopy(matrix_data_stack_entry, matrix_data_stack_entry1,
                        sizeof(LOCAL_REAL_VALUES));

                add_to_matrix_data(matrix_data_stack_entry,
                        &matrix_data_stack_top, &matrix_data_stack);

                add_to_matrix_data(matrix_data_stack_entry1,
                                &(data_extra_item -> stack_top),
                                &(data_extra_item -> stack));

                break;

            case END_OF_MATRIX_LETTER:
                context = (int)Header[4];

#if 0
                if
                (
                    (matrix_data_stack_entry == NULL) ||
                    (context != matrix_data_stack_entry -> context)
                )
                {
                   print("Fatal error: context mismatch in matrix handling.");
                   exit(1);
                }
#endif

#if DEBUG4
                print("END_OF_MATRIX_LETTER...checkpoint #1");
#endif
                break;

            case RHS_DATA_LETTER:
#if DEBUG4
                print("RHS_DATA_LETTER...");
#endif
                /* store the RHS info. */
                bcopy(Header + n_longs, Extra_RHS, n_objs * sizeof(double));
                break;

            case END_OF_DBASE_LETTER:
#if DEBUG4
                print("END_OF_DBASE_LETTER...");
#endif
                done = TRUE;
                break;
            default:
#if DEBUG4
                print("error: unknown letter... %d", Letter_Type);
#endif
                break;
        }

        Zip_free(Letter); Letter = NULL;
    }

#if DEBUG1 | DEBUG9
    print("node(p=%d,q=%d): csparsetest1a -- Checkpoint 2a",
                Grid_Row, Grid_Col);
#endif

    /* initialize local/global statistics: */
    mean_stats = NEW_CLEAR_ITEM(Cstats, NSTATS);
    max_stats  = NEW_CLEAR_ITEM(STATS, NSTATS);
    min_stats  = NEW_CLEAR_ITEM(STATS, NSTATS);

    mean_stats -> g2mlr = Mailer;

    mean_time  = NEW_CLEAR_ITEM(double, NSTATS);
    max_time   = NEW_CLEAR_ITEM(double, NSTATS);
    min_time   = NEW_CLEAR_ITEM(double, NSTATS);

    /* main loop: perform test computations and compute statistics: */
    for (Dump_Count = 0, Counter = 0; Counter < Repeat; Counter++)
    {
#if LED_DISPL
        led_num(0);                     /* top of loop */
#endif

#if DEBUG_STAT_LOOP
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#else
             G2_sync();
             if(G2_pm)
                 print("top of loop Counter = %d; Dump_Count = %d; Repeat= %d",
                         Counter, Dump_Count, Repeat);
#endif
#endif

#if  0
        sizeof_Dmatrix(Csparse_Matrix, &Size);
        print("node(p=%d,q=%d): Calculation Repeat %d, LU info. Size = %d",
                Grid_Row, Grid_Col, Counter, Size);
#endif
        /* set RHS: */
#if DEBUG_STAT_LOOP
G2_sync();
if(G2_pm)
    print("before RHS bcopy, Extra_RHS = %lxX, RHS = %lxX", Extra_RHS, RHS);
#endif

        bcopy(Extra_RHS, RHS, Local_Max_Row * sizeof(double));

#if DEBUG_STAT_LOOP
#if LED_DISPL
        led_num(1);                     /* before mk_matrix */
#endif

G2_sync();
if(G2_pm) print("after RHS bcopy()...before mk_matrix()");
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#endif

        /* fill the matrix: */
        G2_sync(); /* synchronize timing start: */
        Time_Start = clock();

        mk_matrix(Csparse_Matrix, mk_matrix_stack_top, 0, Regen_Flag);


        Time_Stop = clock(); /* local time measurement */

#if DEBUG_STAT_LOOP 
G2_sync();
if(G2_pm) print("after mk_matrix()");
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#endif

        mean_time[FILL_A_STATS] = ((double)(Time_Stop - Time_Start))/1000.0;
        Cupdate_stat_local_data(&mean_stats[FILL_A_STATS], 
                                &mean_time[FILL_A_STATS], 1L);
                                /* update local statistics (for mean time): */

#if DEBUG_STAT_LOOP
if(G2_pm)
    print("after FILL_A_STATS; delta_T = %15.6e", mean_time[FILL_A_STATS]);
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#endif

#if LED_DISPL
        led_num(2);                     /* after mk_matrix */
#endif

        G2_sync(); /* synchronize timing start: */
        Time_Start = clock();

print("before lu_factor_Dmatrix()\n");
        /* factor the matrix (A-mode): */
        lu_factor_Dmatrix(Matrix_LU_Info, NULL, Right_Hand_Side);
print("after lu_factor_Dmatrix()\n");

        Time_Stop = clock(); /* local time measurement: */

        mean_time[FACTOR_A_STATS] = (double)((Time_Stop - Time_Start))/1000.0;
        Cupdate_stat_local_data(&mean_stats[FACTOR_A_STATS],
                                &mean_time[FACTOR_A_STATS], 1L);
                                /* update local statistics (for mean time): */

#if DEBUG_STAT_LOOP
if(G2_pm)
    print("after FACTOR_A_STATS; rank = %d; delta_T = %15.6e",
        Matrix_LU_Info -> rank, mean_time[FACTOR_A_STATS]);
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#endif
        /* do the back-solve: */
#if LED_DISPL
        led_num(3);                     /* after "A" mode factorization */
#endif

        G2_sync(); /* synchronize timing start: */

#if DEBUG7dot1
print("after G2_sync() after FACTOR_A_STATS");
print("Matrix_LU_Info = %lxX, Right_Hand_Side = %lxX, Solution = %lxX",
        Matrix_LU_Info, Right_Hand_Side, Solution);
#endif

        Time_Start = clock();

        back_solve_lu_Dmatrix(Matrix_LU_Info, NULL, 0, Right_Hand_Side,
                                        Solution);

        Time_Stop = clock(); /* local time measurement */

#if DEBUG7dot1
print("after back_solve_lu_Dmatrix()...");
#endif

        mean_time[HALF_SOLVE_STATS] = (double)((Time_Stop - Time_Start))/1000.0;
        Cupdate_stat_local_data(&mean_stats[HALF_SOLVE_STATS],
                                &mean_time[HALF_SOLVE_STATS], 1L);
                                /* update local statistics (for mean time): */

#if DEBUG_STAT_LOOP
if(G2_pm) 
    print("after HALF_SOLVE_STATS; delta_T = %15.6e", mean_time[HALF_SOLVE_STATS]);
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#endif

#if LED_DISPL
        led_num(4);                     /* after half solve */
#endif

        /* set RHS before forward/back solve: */
        bcopy(Extra_RHS, RHS, Local_Max_Row * sizeof(double));

#if DEBUG7dot1
print("before solve_lu_Dmatrix()...");
#endif
        G2_sync();
        Time_Start = clock();

        solve_lu_Dmatrix(Matrix_LU_Info, Right_Hand_Side, Solution);

        Time_Stop  = clock();
#if DEBUG7dot1
print("after solve_lu_Dmatrix()...");
#endif

        mean_time[SOLVE_STATS] = (double)((Time_Stop - Time_Start))/1000.0;
        Cupdate_stat_local_data(&mean_stats[SOLVE_STATS],
                                &mean_time[SOLVE_STATS], 1L);
                                /* update local statistics (for mean time): */


#if DEBUG_STAT_LOOP
if(G2_pm)
    print("after SOLVE_STATS; delta_T = %15.6e", mean_time[SOLVE_STATS]);
#endif
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#if LED_DISPL
        led_num(5);                     /* after full solve */
#endif

        /* minimal reset for B mode: */
        G2_sync(); /* synchronize timing start: */
        Time_Start = clock();

        clear_Dlu_info(Matrix_LU_Info);

        Time_Stop = clock();
        mean_time[CLEAR_LU_STATS] = (double)((Time_Stop - Time_Start))/1000.0;
        Cupdate_stat_local_data(&mean_stats[CLEAR_LU_STATS],
                                &mean_time[CLEAR_LU_STATS], 1L);

#if DEBUG_STAT_LOOP
G2_sync();
if(G2_pm)
    print("after CLEAR_LU_STATS");
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#endif
        /* refill for B-mode factorization: */
        G2_sync(); /* synchronize timing start: */
        Time_Start = clock();

        mk_matrix(Csparse_Matrix, mk_matrix_stack_top, 1, Regen_Flag);

        Time_Stop = clock();

        mean_time[FILL_B_STATS] = (double)((Time_Stop - Time_Start))/1000.0;
        Cupdate_stat_local_data(&mean_stats[FILL_B_STATS],
                                &mean_time[FILL_B_STATS], 1L);
                                /* update local statistics (for mean time): */

#if DEBUG_STAT_LOOP
if(G2_pm)
    print("after FILL_B_STATS; delta_T = %15.6e",mean_time[FILL_B_STATS]);
#endif
        /* set RHS: */
        bcopy(Extra_RHS, RHS, Local_Max_Row * sizeof(double));

#if DEBUG_STAT_LOOP
G2_sync();
if(G2_pm)
    print("after re-bcopy() RHS");
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#endif
#if LED_DISPL
        led_num(6);                     /* before B mode */
#endif

        /* set RHS: */
        /* factor the matrix (B-mode): */
        G2_sync(); /* synchronize timing start: */
        Time_Start = clock();

        lu_factor_Dmatrix(Matrix_LU_Info, NULL, Right_Hand_Side);

        Time_Stop = clock();

        mean_time[FACTOR_B_STATS] = (double)((Time_Stop - Time_Start))/1000.0;
        Cupdate_stat_local_data(&mean_stats[FACTOR_B_STATS],
                                &mean_time[FACTOR_B_STATS], 1L);
                                /* update local statistics (for mean time): */

#if DEBUG_STAT_LOOP
if(G2_pm) 
    print("after FACTOR_B_STATS; delta_T = %15.6e", mean_time[FACTOR_B_STATS]);
#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#endif

        /* reset (==> abandon structure) the sparse matrix: */
        G2_sync(); /* synchronize timing start: */
        Time_Start = clock();
        
        reset_Dlu_info(Matrix_LU_Info);

        Time_Stop = clock();

        mean_time[RESET_LU_STATS] = (double)((Time_Stop - Time_Start))/1000.0;
        Cupdate_stat_local_data(&mean_stats[RESET_LU_STATS],
                                &mean_time[RESET_LU_STATS], 1L);
                                /* update local statistics (for mean time): */

        /* update statistics for max. times: */
#if DEBUG_STAT_LOOP
if(G2_pm)
    print("after RESET_LU_STATS; delta_T = %15.6e", mean_time[RESET_LU_STATS]);
#endif

#if LED_DISPL
        led_num(7);                     /* after "B" mode & reset */
#endif
        
        /* find maxima over ensemble: */
        bcopy(mean_time, max_time, NSTATS * sizeof(double));
        G2_combine(max_time, dmax_co, sizeof(double), NSTATS);

        /* find minima over ensemble: */
        bcopy(mean_time, min_time, NSTATS * sizeof(double));
        G2_combine(min_time, dmin_co, sizeof(double), NSTATS);

#ifdef DEBUG_STATS_IN_MAIN
if(G2_pm)
    for(i = 0; i < NSTATS; i++)
        print("main(): min[%5d] = %15.6g ; max[%5d] = %15.6g",
            i, min_time[i], i, max_time[i]);
#endif

        /* each node has same statistics for these quantities: */
        for(i = 0; i < NSTATS; i++)
        {
            update_stat_data(min_stats + i, min_time + i, 1);
            update_stat_data(max_stats + i, max_time + i, 1);
        }

        /* incremental dumping of results (for safety) */
        if((!((Counter+1) % Dump_to_Host_Freq)) && (Counter < (Repeat-1)))
            dump_stats_to_host(Mailer, Dump_Count++, NSTATS, mean_stats,
                        min_stats, max_stats);
#if DEBUG_STAT_LOOP
G2_sync();
if(G2_pm)
    print("bottom of loop Counter = %d",Counter);

#if DEBUG_STAT_LOOP_MEM_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif
#endif
    }
#if DEBUG_STAT_LOOP
    G2_sync();
    if(G2_pm)
        print("after STAT_LOOP");
#endif

    /* final dump of stats */
    dump_stats_to_host(Mailer,-1,NSTATS,mean_stats, min_stats, max_stats);

#if MEMORY_DISPL
memory_displ(Grid_Row, Grid_Col);
#endif

    /* all done */ 
    Zip_pop();
    zip_exit();
 
#if DEBUG1
    print("node(p=%d,q=%d): csparsetest1a -- done", Grid_Row, Grid_Col);
#endif
}

#define DEBUG_DUMP_STATS_TO_HOST 1
#define DEBUG_DUMP_STATS_TO_HOST_VERBOSE 0
dump_stats_to_host(Mailer,
        dump_count, nstats, mean_stats, min_stats, max_stats)
ZIP_MAILER *Mailer;
long dump_count;
long nstats;
Cstats *mean_stats;
STATS  *min_stats, *max_stats;
{
    int i;
    char *Letter;
    long Letter_Size, n_longs;
    long *Header;
    short *Stats_short;
    long *Stats_long;
    double *Stats_double;

    int Max_Grid_Row, Max_Grid_Col, Grid_Row, Grid_Col;
    int G2_pm;

    CSTATS *global_stats;
    STATS  *the_stats;

    STASIS G2_POBOX        Send_POBox;

#if DEBUG_DUMP_STATS_TO_HOST
    char *subroutine_name = "dump_stats_to_host()";

    print("%s: entry... dump_count = %d, nstats = %ld", subroutine_name,
          dump_count, nstats);
#endif

    Zip_push(Mailer);

    G2_PQ(Max_Grid_Row, Max_Grid_Col);
    G2_pq(Grid_Row, Grid_Col);
 
    G2_pm = (!Grid_Row) && (!Grid_Col);

    Send_POBox.d1 = Grid_Row;
    Send_POBox.d2 = Grid_Col;
 
    /* compute global statistics: */
#if DEBUG_DUMP_STATS_TO_HOST
    print("%s: checkpoint 0\n", subroutine_name);
#endif
    Ccompute_Nstats(mean_stats, nstats);

    /* min and max stats are already based on global quantities,
       just form sigma & sigma_mean for them: */

#if DEBUG_DUMP_STATS_TO_HOST
    print("%s: checkpoint 1\n", subroutine_name);
#endif

    for(i = 0; i < nstats; i++)
    {
        compute_stats(max_stats + i);
        compute_stats(min_stats + i);
    }

    if (G2_pm)
    {
#if DEBUG_DUMP_STATS_TO_HOST
    print("%s: checkpoint 2\n", subroutine_name);
#endif

        /* construct global statistical information for host: */
#if 0
#define NLONGS  6
#define SHORTS_PER_STAT 4
#define LONGS_PER_STAT  (1 + 0 + 1 + 1)
#define REALS_PER_STAT  (2 + 4 + 4 + 4)
#endif

        Letter_Size = NLONGS*sizeof(long)
                      + nstats * (SHORTS_PER_STAT * sizeof(short) +
                                  LONGS_PER_STAT * sizeof(long) +
                                  REALS_PER_STAT * sizeof(double));

        Letter      = Zip_malloc(Letter_Size);
        n_longs     = NLONGS;

        Header      = (long *)Letter;
        Header[0]   = STAT_LETTER;
        Header[1]   = Letter_Size;
        Header[2]   = n_longs;
        Header[3]   = nstats;
        Header[4]   = (long)dump_count;
        Header[5]   = 0L;

        Stats_short  = (short *)(Header + n_longs);

#if DEBUG_DUMP_STATS_TO_HOST
    print("dump_stats_to_host(): checkpoint 2.1\n");
#endif

        for(i = 0; i < nstats; i++)
        {
            /* overall min/max and where achieved: */
            global_stats = &(mean_stats[i].global_stats);

            Stats_short[0]  = global_stats -> min_p;
            Stats_short[1]  = global_stats -> min_q;
            Stats_short[2] = global_stats -> max_p;
            Stats_short[3] = global_stats -> max_q;
#if 0
            ctohs(Stats_short, SHORTS_PER_STAT);
#endif
            Stats_short += SHORTS_PER_STAT;
        }

#if DEBUG_DUMP_STATS_TO_HOST
    print("%s: checkpoint 2.2\n", subroutine_name);
#endif

        Stats_long   = (long *)Stats_short;

        for(i = 0; i < nstats; i++)
        {
            Stats_long[0] = (long)(mean_stats[i].global_stats.stats.n);
            Stats_long[1] = (long)(min_stats[i].n);
            Stats_long[2] = (long)(max_stats[i].n);
#if 0
            ctohl(Stats_long, LONGS_PER_STAT);
#endif
            Stats_long += 3;
        }

#if DEBUG_DUMP_STATS_TO_HOST
    print("%s: checkpoint 2.3\n", subroutine_name);
#endif

        Stats_double = (double *)Stats_long;

        for(i = 0; i < nstats; i++)
        {
            Stats_double[0] = mean_stats[i].global_stats.stats.min_value;
            Stats_double[1] = mean_stats[i].global_stats.stats.max_value;

            /* global mean distribution: */
            Stats_double[2] = mean_stats[i].global_stats.stats.mean;
            Stats_double[3] = mean_stats[i].global_stats.stats.sigma;
            Stats_double[4] = mean_stats[i].global_stats.stats.sigma_mean;
            Stats_double[5] = mean_stats[i].global_stats.stats.sum_squared;

            /* distribution of the min's: */
            Stats_double[6] = min_stats[i].mean;
            Stats_double[7] = min_stats[i].sigma;
            Stats_double[8] = min_stats[i].sigma_mean;
            Stats_double[9] = min_stats[i].sum_squared;

            /* distribution of the max's: */
            Stats_double[10] = max_stats[i].mean;
            Stats_double[11] = max_stats[i].sigma;
            Stats_double[12] = max_stats[i].sigma_mean;
            Stats_double[13] = max_stats[i].sum_squared;

#if 0
            ctohd(Stats_double, REALS_PER_STAT);
#endif
            Stats_double += REALS_PER_STAT;
        }

#if DEBUG_DUMP_STATS_TO_HOST
    print("%s: checkpoint 2.4\n", subroutine_name);
#endif

        G2_send(Letter, G2_HOST_p, G2_HOST_q);
    }

    Zip_pop();

#if DEBUG_DUMP_STATS_TO_HOST
    print("%s: exiting...", subroutine_name);
#endif
} 

add_to_mk_matrix(matrix_entry, mk_matrix_stack_top, mk_matrix_stack)
MK_MATRIX_ENTRY *matrix_entry;
MK_MATRIX_ENTRY **mk_matrix_stack_top, **mk_matrix_stack;
{
    if(*mk_matrix_stack_top == NULL)
    {
        *mk_matrix_stack_top = *mk_matrix_stack = matrix_entry;
    }
    else
    {
        (*mk_matrix_stack) -> next = matrix_entry;
        *mk_matrix_stack           = matrix_entry;
    }
    matrix_entry -> next = NULL;

}

add_to_matrix_data(matrix_data_stack_entry, matrix_data_stack_top,
                   matrix_data_stack)
LOCAL_REAL_VALUES *matrix_data_stack_entry;
LOCAL_REAL_VALUES **matrix_data_stack_top, **matrix_data_stack;
{
    if(matrix_data_stack_top == NULL)
    {
        *matrix_data_stack_top = *matrix_data_stack =
            matrix_data_stack_entry;
    }
    else
    {
        (*matrix_data_stack) -> next = matrix_data_stack_entry;
        *matrix_data_stack           = matrix_data_stack_entry;
    }

    matrix_data_stack_entry -> next     = NULL;
}

mk_matrix(Csparse_Matrix, mk_matrix_stack_top, mode, regen)
Dmatrix *Csparse_Matrix;
MK_MATRIX_ENTRY *mk_matrix_stack_top;
int mode;
int regen;
{
    int first = 1;
    MK_MATRIX_ENTRY *entry EQ_NULL;
    SBAND_EXTRA *sband EQ_NULL;

    for(entry = mk_matrix_stack_top; entry != NULL; entry = entry -> next)
    {
        switch(entry -> gen_class)
        {
            case GEN_CLASS_DATA:        /* will use both mode, regen */
                /* not supported for now */
                print("mk_matrix() GEN_CLASS_DATA unsupported for now...");
                break;
            case GEN_CLASS_SBAND:
                sband = (SBAND_EXTRA *)(entry -> extra);
                if(!regen)
                {
                    if(!(sband -> init)) /* generate the first time... */
                    {
                        /* mode == 1 here is an error... */

                        gen_sband_random_Dmatrix_data(Csparse_Matrix,
                            &(sband -> stack_top), &(sband -> stack), sband);

                        sband -> init = 1;
                    }

                    fill_Smatrix(Csparse_Matrix -> Local_Matrix,
                                        sband -> stack_top, mode, first);
                }
                else
                {
#if 0
                    if(sband -> stack_top != NULL)
                    {
                        free_local_values(sband -> stack_top);
                        sband -> stack_top = sband -> stack = NULL;
                    }
#endif

                    /* for mode == 1, structure retained;
                        fill untouched -- shouldn't hurt??? */
                    gen_sband_random_Dmatrix(Csparse_Matrix, sband, mode,
                                                                        first);
                }

                break;
            default:
                print("mk_matrix() unknown gen_class %d",entry -> gen_class);
                break;

        } /* end switch */

        first = 0;

    } /* end for() */
}

free_local_values(stack_top)
LOCAL_REAL_VALUES *stack_top;
{
    LOCAL_REAL_VALUES *stack, *stack_next;

    for(stack = stack_top; stack != NULL; stack = stack_next)
    {
        stack_next = stack -> next;
        Free(stack -> values);
        Free(stack);
    }
}

int imin_max_avg_where_co(buf1, buf2, size, nitems)
int *buf1, *buf2;
int size;
int nitems;
{
    int i;

    if(size != 10*sizeof(int))
    {
        print("min_max_avg_where_co(): illegal size");
        exit(0);
    }

    for(i = 0; i < nitems; i++, buf1 += 10, buf2 += 10)
    {
        /* apply a totally associative min test: */

        if
        (   (buf1[0] > buf2[0]) ||
            (
                        (buf1[0] == buf2[0]) && 
                (
                                (buf1[1] > buf2[1]) ||
                    ((buf1[1] == buf2[1]) && (buf1[2] > buf2[1]))
                )
            )
        )
        {
            if(buf1[0] == buf2[0])
                buf1[3] += buf2[3];     /* multiplicity increase */

            buf1[0] = buf2[0];
            buf1[1] = buf2[1];
            buf1[2] = buf2[2];
        }

        /* apply a totally associative max test: */
        if
        (   (buf1[4] < buf2[4]) ||
            (
                        (buf1[4] == buf2[4]) && 
                (
                                (buf1[5] > buf2[5]) ||
                    ((buf1[5] == buf2[5]) && (buf1[6] > buf2[6]))
                )
            )
        )
        {
            if(buf1[4] == buf2[4])
                buf1[7] += buf2[7];     /* multiplicity increase */

            buf1[4] = buf2[4];
            buf1[5] = buf2[5];
            buf1[6] = buf2[6];
        }

        /* sum of quantities: */
        buf1[8] += buf2[8];
        buf1[9] += buf2[9];

    }
}

#if LED_DISPL
/*+
 * NAME: led_num
 * PURPOSE: show a binary number on the leds. E.g. led_num(5) will light up
 *          the first and third leds.
 * ARGUMENTS:
 *      num - number to show.
 * RETURN VALUE: none
 *-
 * courtesy of sharon brunett, origin unknown (symult???) {AS}
 */

led_num(num)
int num;
{
    int i;                    /* led number */
    for( i = 0; i < 4; i++ ){
      if( num & (1 << i) )
          led(1,i);
      else
          led(0,i);
    }
}
#endif

memory_displ(Grid_Row, Grid_Col)
int Grid_Row, Grid_Col;
{
    int my_free_memory;
    int imin_max_avg_where_co();
    int min_max_avg[10];
    int G2_pm;

    G2_pm = (!Grid_Row) && (!Grid_Col);

        my_free_memory = SIMPLE_RLIM;
        min_max_avg[0] = my_free_memory;        /* min slot */
        min_max_avg[1] = Grid_Row;
        min_max_avg[2] = Grid_Col;
        min_max_avg[3] = 1;                     /* multiplicity of min */

        min_max_avg[4] = my_free_memory;        /* max slot */
        min_max_avg[5] = Grid_Row;
        min_max_avg[6] = Grid_Col;
        min_max_avg[7] = 1;                     /* multiplicity of max */

        min_max_avg[8] = my_free_memory;        /* sum */
        min_max_avg[9] = 1;                     /* count */

        G2_combine(min_max_avg, imin_max_avg_where_co, 10*sizeof(int), 1);

        if(G2_pm)
        {
#if 0
            print("top of loop Counter = %d; Dump_Count = %d; Repeat= %d; memory:", Counter, Dump_Count, Repeat);
#endif
            print("min = %5d, @(p=%d,q=%d) %dx; max = %5d @(p=%d,q=%d) %dx; ",
                min_max_avg[0], min_max_avg[1], min_max_avg[2], min_max_avg[3],
                min_max_avg[4], min_max_avg[5], min_max_avg[6], min_max_avg[7]);
            print("'avg' memory = %g",
                (double)min_max_avg[8] / (double)min_max_avg[9]);
        }
}
